from typing import Optional

from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class CodeVersionChangedCondition(AutomationCondition):
    label: Optional[str] = None

    @property
    def description(self) -> str:
        return "Asset code version changed since previous tick"

    @property
    def name(self) -> str:
        return "code_version_changed"

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        previous_code_version = context.cursor
        current_code_version = context.asset_graph.get(context.asset_key).code_version
        if previous_code_version is None or previous_code_version == current_code_version:
            true_slice = context.get_empty_slice()
        else:
            true_slice = context.candidate_slice

        return AutomationResult(context, true_slice, cursor=current_code_version)



import datetime
from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Mapping, Optional, Sequence

from typing_extensions import Self

import dagster._check as check
from dagster._annotations import experimental, public
from dagster._core.asset_graph_view.asset_graph_view import AssetSlice
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetSubsetWithMetadata,
    AutomationConditionCursor,
    AutomationConditionEvaluation,
    AutomationConditionNodeCursor,
    AutomationConditionNodeSnapshot,
    AutomationConditionSnapshot,
    get_serializable_candidate_subset,
)
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._core.definitions.time_window_partitions import BaseTimeWindowPartitionsSubset
from dagster._record import copy
from dagster._serdes.serdes import is_whitelisted_for_serdes_object
from dagster._time import get_current_timestamp
from dagster._utils.security import non_secure_md5_hash_str
from dagster._utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
    from dagster._core.definitions.declarative_automation.automation_context import (
        AutomationContext,
    )
    from dagster._core.definitions.declarative_automation.operands import (
        CodeVersionChangedCondition,
        CronTickPassedCondition,
        FailedAutomationCondition,
        InLatestTimeWindowCondition,
        InProgressAutomationCondition,
        MissingAutomationCondition,
        NewlyRequestedCondition,
        NewlyUpdatedCondition,
        WillBeRequestedCondition,
    )
    from dagster._core.definitions.declarative_automation.operators import (
        AllDepsCondition,
        AndAutomationCondition,
        AnyDepsCondition,
        AnyDownstreamConditionsCondition,
        NewlyTrueCondition,
        NotAutomationCondition,
        OrAutomationCondition,
        SinceCondition,
    )


class AutomationCondition(ABC):
    """An AutomationCondition represents a condition of an asset that impacts whether it should be
    automatically executed. For example, you can have a condition which becomes true whenever the
    code version of the asset is changed, or whenever an upstream dependency is updated.

    .. code-block:: python

        from dagster import AutomationCondition, asset

        @asset(automation_condition=AutomationCondition.on_cron("0 0 * * *"))
        def my_asset(): ...

    AutomationConditions may be combined together into expressions using a variety of operators.

    .. code-block:: python

        from dagster import AssetSelection, AutomationCondition, asset

        # any dependencies from the "important" group are missing
        any_important_deps_missing = AutomationCondition.any_deps_match(
            AutomationCondition.missing(),
        ).allow(AssetSelection.groups("important"))

        # there is a new code version for this asset since the last time it was requested
        new_code_version = AutomationCondition.code_version_changed().since(
            AutomationCondition.newly_requested()
        )

        # there is a new code version and no important dependencies are missing
        my_condition = new_code_version & ~any_important_deps_missing

        @asset(automation_condition=my_condition)
        def my_asset(): ...

    """

    @property
    def requires_cursor(self) -> bool:
        return True

    @property
    def children(self) -> Sequence["AutomationCondition"]:
        return []

    @property
    def description(self) -> str:
        """Human-readable description of when this condition is true."""
        return ""

    @property
    def label(self) -> Optional[str]:
        """User-provided label subjectively describing the purpose of this condition in the broader evaluation tree."""
        return None

    @property
    def name(self) -> str:
        """Formal name of this specific condition, generally aligning with its static constructor."""
        return self.__class__.__name__

    def get_node_snapshot(self, unique_id: str) -> AutomationConditionNodeSnapshot:
        """Returns a snapshot of this condition that can be used for serialization."""
        return AutomationConditionNodeSnapshot(
            class_name=self.__class__.__name__,
            description=self.description,
            unique_id=unique_id,
            label=self.label,
            name=self.name,
        )

    def get_snapshot(
        self, *, parent_unique_id: Optional[str] = None, index: Optional[int] = None
    ) -> AutomationConditionSnapshot:
        """Returns a serializable snapshot of the entire AutomationCondition tree."""
        unique_id = self.get_unique_id(parent_unique_id=parent_unique_id, index=index)
        node_snapshot = self.get_node_snapshot(unique_id)
        children = [
            child.get_snapshot(parent_unique_id=unique_id, index=i)
            for (i, child) in enumerate(self.children)
        ]
        return AutomationConditionSnapshot(node_snapshot=node_snapshot, children=children)

    def get_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[int]) -> str:
        """Returns a unique identifier for this condition within the broader condition tree."""
        parts = [str(parent_unique_id), str(index), self.__class__.__name__, self.description]
        return non_secure_md5_hash_str("".join(parts).encode())

    def get_hash(
        self, *, parent_unique_id: Optional[str] = None, index: Optional[int] = None
    ) -> int:
        """Generates a hash based off of the unique ids of all children."""
        unique_id = self.get_unique_id(parent_unique_id=parent_unique_id, index=index)
        hashes = [hash(unique_id)]
        for i, child in enumerate(self.children):
            hashes.append(child.get_hash(parent_unique_id=unique_id, index=i))
        return hash(tuple(hashes))

    def __hash__(self) -> int:
        return self.get_hash()

    @property
    def has_rule_condition(self) -> bool:
        from dagster._core.definitions.declarative_automation.legacy import RuleCondition

        if isinstance(self, RuleCondition):
            return True
        return any(child.has_rule_condition for child in self.children)

    @property
    def is_serializable(self) -> bool:
        if not is_whitelisted_for_serdes_object(self):
            return False
        return all(child.is_serializable for child in self.children)

    def as_auto_materialize_policy(self) -> "AutoMaterializePolicy":
        """Returns an AutoMaterializePolicy which contains this condition."""
        from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy

        return AutoMaterializePolicy.from_automation_condition(self)

    @abstractmethod
    def evaluate(self, context: "AutomationContext") -> "AutomationResult":
        raise NotImplementedError()

    def __and__(self, other: "AutomationCondition") -> "AndAutomationCondition":
        from dagster._core.definitions.declarative_automation.operators import (
            AndAutomationCondition,
        )

        # group AndAutomationConditions together
        if isinstance(self, AndAutomationCondition):
            return AndAutomationCondition(operands=[*self.operands, other])
        return AndAutomationCondition(operands=[self, other])

    def __or__(self, other: "AutomationCondition") -> "OrAutomationCondition":
        from dagster._core.definitions.declarative_automation.operators import OrAutomationCondition

        # group OrAutomationConditions together
        if isinstance(self, OrAutomationCondition):
            return OrAutomationCondition(operands=[*self.operands, other])
        return OrAutomationCondition(operands=[self, other])

    def __invert__(self) -> "NotAutomationCondition":
        from dagster._core.definitions.declarative_automation.operators import (
            NotAutomationCondition,
        )

        return NotAutomationCondition(operand=self)

    @public
    def with_label(self, label: Optional[str]) -> Self:
        """Returns a copy of this AutomationCondition with a human-readable label."""
        return copy(self, label=label)

    def since(self, reset_condition: "AutomationCondition") -> "SinceCondition":
        """Returns an AutomationCondition that is true if this condition has become true since the
        last time the reference condition became true.
        """
        from dagster._core.definitions.declarative_automation.operators import SinceCondition

        return SinceCondition(trigger_condition=self, reset_condition=reset_condition)

    def since_last_handled(self) -> "SinceCondition":
        """Returns an AutomationCondition that is true if this condition has become true since the
        last time this asset partition was requested or updated.
        """
        with disable_dagster_warnings():
            return self.since(
                (
                    AutomationCondition.newly_requested() | AutomationCondition.newly_updated()
                ).with_label("handled")
            )

    def newly_true(self) -> "NewlyTrueCondition":
        """Returns an AutomationCondition that is true only on the tick that this condition goes
        from false to true for a given asset partition.
        """
        from dagster._core.definitions.declarative_automation.operators import NewlyTrueCondition

        return NewlyTrueCondition(operand=self)

    @public
    @experimental
    @staticmethod
    def any_deps_match(condition: "AutomationCondition") -> "AnyDepsCondition":
        """Returns an AutomationCondition that is true for an asset partition if at least one partition
        of any of its dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's dependencies.
        """
        from dagster._core.definitions.declarative_automation.operators import AnyDepsCondition

        return AnyDepsCondition(operand=condition)

    @public
    @experimental
    @staticmethod
    def all_deps_match(condition: "AutomationCondition") -> "AllDepsCondition":
        """Returns an AutomationCondition that is true for an asset partition if at least one partition
        of all of its dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's dependencies.
        """
        from dagster._core.definitions.declarative_automation.operators import AllDepsCondition

        return AllDepsCondition(operand=condition)

    @public
    @experimental
    @staticmethod
    def missing() -> "MissingAutomationCondition":
        """Returns an AutomationCondition that is true for an asset partition if it has never been
        materialized or observed.
        """
        from dagster._core.definitions.declarative_automation.operands import (
            MissingAutomationCondition,
        )

        return MissingAutomationCondition()

    @public
    @experimental
    @staticmethod
    def in_progress() -> "InProgressAutomationCondition":
        """Returns an AutomationCondition that is true for an asset partition if it is part of an in-progress run."""
        from dagster._core.definitions.declarative_automation.operands import (
            InProgressAutomationCondition,
        )

        return InProgressAutomationCondition()

    @public
    @experimental
    @staticmethod
    def failed() -> "FailedAutomationCondition":
        """Returns an AutomationCondition that is true for an asset partition if its latest run failed."""
        from dagster._core.definitions.declarative_automation.operands import (
            FailedAutomationCondition,
        )

        return FailedAutomationCondition()

    @public
    @experimental
    @staticmethod
    def in_latest_time_window(
        lookback_delta: Optional[datetime.timedelta] = None,
    ) -> "InLatestTimeWindowCondition":
        """Returns an AutomationCondition that is true for an asset partition when it is within the latest
        time window.

        Args:
            lookback_delta (Optional, datetime.timedelta): If provided, the condition will
                return all partitions within the provided delta of the end of the latest time window.
                For example, if this is used on a daily-partitioned asset with a lookback_delta of
                48 hours, this will return the latest two partitions.
        """
        from dagster._core.definitions.declarative_automation.operands import (
            InLatestTimeWindowCondition,
        )

        return InLatestTimeWindowCondition.from_lookback_delta(lookback_delta)

    @public
    @experimental
    @staticmethod
    def will_be_requested() -> "WillBeRequestedCondition":
        """Returns an AutomationCondition that is true for an asset partition if it will be requested this tick."""
        from dagster._core.definitions.declarative_automation.operands import (
            WillBeRequestedCondition,
        )

        return WillBeRequestedCondition()

    @public
    @experimental
    @staticmethod
    def newly_updated() -> "NewlyUpdatedCondition":
        """Returns an AutomationCondition that is true for an asset partition if it has been updated since the previous tick."""
        from dagster._core.definitions.declarative_automation.operands import NewlyUpdatedCondition

        return NewlyUpdatedCondition()

    @public
    @experimental
    @staticmethod
    def newly_requested() -> "NewlyRequestedCondition":
        """Returns an AutomationCondition that is true for an asset partition if it was requested on the previous tick."""
        from dagster._core.definitions.declarative_automation.operands import (
            NewlyRequestedCondition,
        )

        return NewlyRequestedCondition()

    @public
    @experimental
    @staticmethod
    def code_version_changed() -> "CodeVersionChangedCondition":
        """Returns an AutomationCondition that is true for an asset partition if its asset's code
        version has been changed since the previous tick.
        """
        from dagster._core.definitions.declarative_automation.operands import (
            CodeVersionChangedCondition,
        )

        return CodeVersionChangedCondition()

    @public
    @experimental
    @staticmethod
    def cron_tick_passed(
        cron_schedule: str, cron_timezone: str = "UTC"
    ) -> "CronTickPassedCondition":
        """Returns an AutomationCondition that is true for all asset partitions whenever a cron tick of the provided schedule is passed."""
        from dagster._core.definitions.declarative_automation.operands import (
            CronTickPassedCondition,
        )

        return CronTickPassedCondition(
            cron_schedule=cron_schedule, cron_timezone=cron_timezone
        ).with_label(f"cron_tick_passed({cron_schedule}, {cron_timezone})")

    @experimental
    @staticmethod
    def newly_missing() -> "AutomationCondition":
        """Returns an AutomationCondition that is true on the tick that an asset partition becomes missing."""
        with disable_dagster_warnings():
            return AutomationCondition.missing().newly_true().with_label("newly_missing")

    @experimental
    @staticmethod
    def any_deps_updated() -> "AnyDepsCondition":
        """Returns an AutomationCondition that is true for any asset partition with at least one
        dependency that has updated since the previous tick, or will be requested on this tick.
        """
        with disable_dagster_warnings():
            return AutomationCondition.any_deps_match(
                AutomationCondition.newly_updated() | AutomationCondition.will_be_requested()
            ).with_label("any_deps_updated")

    @experimental
    @staticmethod
    def any_deps_missing() -> "AnyDepsCondition":
        """Returns an AutomationCondition that is true for any asset partition with at least one
        dependency that is missing, and will not be requested on this tick.
        """
        with disable_dagster_warnings():
            return AutomationCondition.any_deps_match(
                AutomationCondition.missing() & ~AutomationCondition.will_be_requested()
            ).with_label("any_deps_missing")

    @experimental
    @staticmethod
    def any_deps_in_progress() -> "AnyDepsCondition":
        """Returns an AutomationCondition that is true for any asset partition with at least one
        dependency that is in progress.
        """
        with disable_dagster_warnings():
            return AutomationCondition.any_deps_match(AutomationCondition.in_progress()).with_label(
                "any_deps_in_progress"
            )

    @experimental
    @staticmethod
    def all_deps_updated_since_cron(
        cron_schedule: str, cron_timezone: str = "UTC"
    ) -> "AllDepsCondition":
        """Returns an AutomatonCondition that is true for any asset partition where all of its
        dependencies have updated since the latest tick of the provided cron schedule.
        """
        with disable_dagster_warnings():
            return AutomationCondition.all_deps_match(
                AutomationCondition.newly_updated().since(
                    AutomationCondition.cron_tick_passed(cron_schedule, cron_timezone)
                )
                | AutomationCondition.will_be_requested()
            ).with_label(f"all_deps_updated_since_cron({cron_schedule}, {cron_timezone})")

    @public
    @experimental
    @staticmethod
    def eager() -> "AutomationCondition":
        """Returns an AutomationCondition which will cause missing asset partitions to be
        materialized, and will materialize asset partitions whenever their parents are updated.

        For time partitioned assets, only the latest time partition will be considered.

        This will never evaluate to true if the asset has any upstream partitions which are missing
        or part of an in progress run, and will never evaluate to true if the provided asset partition
        is already part of an in progress run.
        """
        with disable_dagster_warnings():
            return (
                AutomationCondition.in_latest_time_window()
                & (
                    AutomationCondition.newly_missing() | AutomationCondition.any_deps_updated()
                ).since_last_handled()
                & ~AutomationCondition.any_deps_missing()
                & ~AutomationCondition.any_deps_in_progress()
                & ~AutomationCondition.in_progress()
            ).with_label("eager")

    @public
    @experimental
    @staticmethod
    def on_cron(cron_schedule: str, cron_timezone: str = "UTC") -> "AutomationCondition":
        """Returns an AutomationCondition which will cause asset partitions to be materialized
        on a given cron schedule, after all of their dependencies have been updated since the latest
        tick of that cron schedule.

        For time partitioned assets, only the latest time partition will be considered.
        """
        with disable_dagster_warnings():
            return (
                AutomationCondition.in_latest_time_window()
                & AutomationCondition.cron_tick_passed(
                    cron_schedule, cron_timezone
                ).since_last_handled()
                & AutomationCondition.all_deps_updated_since_cron(cron_schedule, cron_timezone)
            ).with_label(f"on_cron({cron_schedule}, {cron_timezone})")

    @public
    @experimental
    @staticmethod
    def any_downstream_conditions() -> "AnyDownstreamConditionsCondition":
        """Returns an AutomationCondition which represents the union of all distinct downstream conditions."""
        from dagster._core.definitions.declarative_automation.operators import (
            AnyDownstreamConditionsCondition,
        )

        return AnyDownstreamConditionsCondition()


class AutomationResult:
    """The result of evaluating an AutomationCondition."""

    def __init__(
        self,
        context: "AutomationContext",
        true_slice: AssetSlice,
        cursor: Optional[str] = None,
        child_results: Optional[Sequence["AutomationResult"]] = None,
        **kwargs,
    ):
        from dagster._core.definitions.declarative_automation.automation_context import (
            AutomationContext,
        )

        self._context = check.inst_param(context, "context", AutomationContext)
        self._true_slice = check.inst_param(true_slice, "true_slice", AssetSlice)
        self._child_results = check.opt_sequence_param(
            child_results, "child_results", of_type=AutomationResult
        )

        self._start_timestamp = context.create_time.timestamp()
        self._end_timestamp = get_current_timestamp()

        # hidden_param which should only be set by legacy RuleConditions
        self._subsets_with_metadata = check.opt_sequence_param(
            kwargs.get("subsets_with_metadata"), "subsets_with_metadata", AssetSubsetWithMetadata
        )

        # hidden_param which should only be set by builtin conditions which require high performance
        # in their serdes layer
        structured_cursor = kwargs.get("structured_cursor")
        invalid_hidden_params = set(kwargs.keys()) - {"subsets_with_metadata", "structured_cursor"}
        check.param_invariant(
            not invalid_hidden_params, "kwargs", f"Invalid hidden params: {invalid_hidden_params}"
        )
        check.param_invariant(
            not (cursor and structured_cursor),
            "structured_cursor",
            "Cannot provide both cursor and structured_cursor.",
        )
        self._extra_state = check.opt_str_param(cursor, "cursor") or structured_cursor

        # used to enable the evaluator class to modify the evaluation in some edge cases
        self._serializable_evaluation_override: Optional[AutomationConditionEvaluation] = None

    @property
    def asset_key(self) -> AssetKey:
        return self._true_slice.asset_key

    @property
    def true_slice(self) -> AssetSlice:
        return self._true_slice

    @property
    def true_subset(self) -> AssetSubset:
        return self.true_slice.convert_to_valid_asset_subset()

    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp

    @property
    def end_timestamp(self) -> float:
        return self._end_timestamp

    @property
    def child_results(self) -> Sequence["AutomationResult"]:
        return self._child_results

    @property
    def condition(self) -> AutomationCondition:
        return self._context.condition

    @property
    def condition_unique_id(self) -> str:
        return self._context.condition_unique_id

    @cached_property
    def value_hash(self) -> str:
        """An identifier for the contents of this AutomationResult. This will be identical for
        results with identical values, allowing us to avoid storing redundant information.
        """
        components: Sequence[str] = [
            self.condition_unique_id,
            self.condition.description,
            _compute_subset_value_str(self.true_subset),
            _compute_subset_value_str(
                self._context.candidate_slice.convert_to_valid_asset_subset()
            ),
            *(_compute_subset_with_metadata_value_str(swm) for swm in self._subsets_with_metadata),
            *(child_result.value_hash for child_result in self._child_results),
        ]
        return non_secure_md5_hash_str("".join(components).encode("utf-8"))

    @cached_property
    def node_cursor(self) -> Optional[AutomationConditionNodeCursor]:
        """Cursor value storing information about this specific evaluation node, if required."""
        if not self.condition.requires_cursor:
            return None
        return AutomationConditionNodeCursor(
            true_subset=self.true_subset,
            candidate_subset=get_serializable_candidate_subset(
                self._context.candidate_slice.convert_to_valid_asset_subset()
            ),
            subsets_with_metadata=self._subsets_with_metadata,
            extra_state=self._extra_state,
        )

    @cached_property
    def _serializable_evaluation(self) -> AutomationConditionEvaluation:
        return AutomationConditionEvaluation(
            condition_snapshot=self.condition.get_node_snapshot(self.condition_unique_id),
            true_subset=self.true_subset,
            candidate_subset=get_serializable_candidate_subset(
                self._context.candidate_slice.convert_to_valid_asset_subset()
            ),
            subsets_with_metadata=self._subsets_with_metadata,
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            child_evaluations=[
                child_result.serializable_evaluation for child_result in self._child_results
            ],
        )

    @property
    def serializable_evaluation(self) -> AutomationConditionEvaluation:
        """Serializable representation of the evaluation of this condition."""
        return self._serializable_evaluation_override or self._serializable_evaluation

    def set_internal_serializable_evaluation_override(
        self, override: AutomationConditionEvaluation
    ) -> None:
        """Internal method for handling edge cases in which the serializable evaluation must be
        updated after evaluation completes.
        """
        self._serializable_evaluation_override = override

    def get_child_node_cursors(self) -> Mapping[str, AutomationConditionNodeCursor]:
        node_cursors = {self.condition_unique_id: self.node_cursor} if self.node_cursor else {}
        for child_result in self._child_results:
            node_cursors.update(child_result.get_child_node_cursors())
        return node_cursors

    def get_new_cursor(self) -> AutomationConditionCursor:
        return AutomationConditionCursor(
            previous_requested_subset=self.serializable_evaluation.true_subset,
            effective_timestamp=self._context.evaluation_time.timestamp(),
            last_event_id=self._context.max_storage_id,
            node_cursors_by_unique_id=self.get_child_node_cursors(),
            result_value_hash=self.value_hash,
        )


def _compute_subset_value_str(subset: AssetSubset) -> str:
    """Computes a unique string representing a given AssetSubsets. This string will be equal for
    equivalent AssetSubsets.
    """
    if isinstance(subset.value, bool):
        return str(subset.value)
    elif isinstance(subset.value, AllPartitionsSubset):
        return AllPartitionsSubset.__name__
    elif isinstance(subset.value, BaseTimeWindowPartitionsSubset):
        return str(
            [
                (tw.start.timestamp(), tw.end.timestamp())
                for tw in sorted(subset.value.included_time_windows)
            ]
        )
    else:
        return str(list(sorted(subset.asset_partitions)))


def _compute_subset_with_metadata_value_str(subset_with_metadata: AssetSubsetWithMetadata):
    return _compute_subset_value_str(subset_with_metadata.subset) + str(
        sorted(subset_with_metadata.frozen_metadata)
    )



from typing import AbstractSet, Optional, Sequence

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._record import record
from dagster._serdes.serdes import whitelist_for_serdes


@record
class DownstreamConditionWrapperCondition(AutomationCondition):
    """Wrapper object which evaluates a condition against a dependency and returns a subset
    representing the subset of downstream asset which has at least one parent which evaluated to
    True.
    """

    downstream_keys: Sequence[AssetKey]
    operand: AutomationCondition
    label: Optional[str] = None

    @property
    def description(self) -> str:
        return ",".join(key.to_user_string() for key in self.downstream_keys)

    @property
    def children(self) -> Sequence[AutomationCondition]:
        return [self.operand]

    @property
    def requires_cursor(self) -> bool:
        return False

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        child_result = self.operand.evaluate(
            context.for_child_condition(
                child_condition=self.operand, child_index=0, candidate_slice=context.candidate_slice
            )
        )
        return AutomationResult(
            context=context, true_slice=child_result.true_slice, child_results=[child_result]
        )


@whitelist_for_serdes
@record
class AnyDownstreamConditionsCondition(AutomationCondition):
    label: Optional[str] = None

    @property
    def description(self) -> str:
        return "Any downstream conditions"

    @property
    def name(self) -> str:
        return "ANY_DOWNSTREAM_CONDITIONS"

    @property
    def requires_cursor(self) -> bool:
        return False

    def _get_ignored_conditions(
        self, context: AutomationContext
    ) -> AbstractSet[AutomationCondition]:
        """To avoid infinite recursion, we do not expand conditions which are already part of the
        evaluation hierarchy.
        """
        ignored_conditions = {context.condition}
        while context.parent_context is not None:
            context = context.parent_context
            ignored_conditions.add(context.condition)
        return ignored_conditions

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        ignored_conditions = self._get_ignored_conditions(context)
        downstream_conditions = context.asset_graph.get_downstream_automation_conditions(
            asset_key=context.asset_key
        )

        true_slice = context.get_empty_slice()
        child_results = []
        for i, (downstream_condition, asset_keys) in enumerate(
            sorted(downstream_conditions.items(), key=lambda x: sorted(x[1]))
        ):
            if downstream_condition in ignored_conditions:
                continue
            child_condition = DownstreamConditionWrapperCondition(
                downstream_keys=list(sorted(asset_keys)), operand=downstream_condition
            )
            child_context = context.for_child_condition(
                child_condition=child_condition,
                child_index=i,
                candidate_slice=context.candidate_slice,
            )
            child_result = child_condition.evaluate(child_context)

            child_results.append(child_result)
            true_slice = true_slice.compute_union(child_result.true_slice)

        return AutomationResult(context=context, true_slice=true_slice, child_results=child_results)



from dagster import Definitions, load_assets_from_modules

from CODE_LOCATION_NAME_PLACEHOLDER import assets  # type: ignore

all_assets = load_assets_from_modules([assets])

defs = Definitions(
    assets=all_assets,
)



from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterable, Mapping, NamedTuple, Optional, Sequence, Set

import dagster._check as check
from dagster._annotations import PublicAttr, experimental_param
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.definitions.utils import (
    resolve_automation_condition,
    validate_asset_owner,
    validate_group_name,
    validate_tags_strict,
)
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.storage.tags import KIND_PREFIX
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.internal_init import IHasInternalInit

if TYPE_CHECKING:
    from dagster._core.definitions.asset_dep import AssetDep, CoercibleToAssetDep

# SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE lives on the metadata of an asset
# (which currently ends up on the Output associated with the asset key)
# whih encodes the execution type the of asset. "Unexecutable" assets are assets
# that cannot be materialized in Dagster, but can have events in the event
# log keyed off of them, making Dagster usable as a observability and lineage tool
# for externally materialized assets.
SYSTEM_METADATA_KEY_ASSET_EXECUTION_TYPE = "dagster/asset_execution_type"


# SYSTEM_METADATA_KEY_IO_MANAGER_KEY lives on the metadata of an asset without a node def and
# determines the io_manager_key that can be used to load it. This is necessary because IO manager
# keys are otherwise encoded inside OutputDefinitions within NodeDefinitions.
SYSTEM_METADATA_KEY_IO_MANAGER_KEY = "dagster/io_manager_key"

# SYSTEM_METADATA_KEY_AUTO_OBSERVE_INTERVAL_MINUTES lives on the metadata of
# external assets resulting from a source asset conversion. It contains the
# `auto_observe_interval_minutes` value from the source asset and is consulted
# in the auto-materialize daemon. It should eventually be eliminated in favor
# of an implementation of `auto_observe_interval_minutes` in terms of
# `AutoMaterializeRule`.
SYSTEM_METADATA_KEY_AUTO_OBSERVE_INTERVAL_MINUTES = "dagster/auto_observe_interval_minutes"

# SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET lives on the metadata of external assets that are
# created for undefined but referenced assets during asset graph normalization. For example, in the
# below definitions, `foo` is referenced by upstream `bar` but has no corresponding definition:
#
#
#     @asset(deps=["foo"])
#     def bar(context: AssetExecutionContext):
#         ...
#
#     defs=Definitions(assets=[bar])
#
# During normalization we create a "stub" definition for `foo` and attach this metadata to it.
SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET = "dagster/auto_created_stub_asset"


@whitelist_for_serdes
class AssetExecutionType(Enum):
    OBSERVATION = "OBSERVATION"
    UNEXECUTABLE = "UNEXECUTABLE"
    MATERIALIZATION = "MATERIALIZATION"


@experimental_param(param="owners")
@experimental_param(param="tags")
class AssetSpec(
    NamedTuple(
        "_AssetSpec",
        [
            ("key", PublicAttr[AssetKey]),
            ("deps", PublicAttr[Iterable["AssetDep"]]),
            ("description", PublicAttr[Optional[str]]),
            ("metadata", PublicAttr[Mapping[str, Any]]),
            ("group_name", PublicAttr[Optional[str]]),
            ("skippable", PublicAttr[bool]),
            ("code_version", PublicAttr[Optional[str]]),
            ("freshness_policy", PublicAttr[Optional[FreshnessPolicy]]),
            ("automation_condition", PublicAttr[Optional[AutomationCondition]]),
            ("owners", PublicAttr[Sequence[str]]),
            ("tags", PublicAttr[Mapping[str, str]]),
            ("partitions_def", PublicAttr[Optional[PartitionsDefinition]]),
        ],
    ),
    IHasInternalInit,
):
    """Specifies the core attributes of an asset, except for the function that materializes or
    observes it.

    An asset spec plus any materialization or observation function for the asset constitutes an
    "asset definition".

    Attributes:
        key (AssetKey): The unique identifier for this asset.
        deps (Optional[AbstractSet[AssetKey]]): The asset keys for the upstream assets that
            materializing this asset depends on.
        description (Optional[str]): Human-readable description of this asset.
        metadata (Optional[Dict[str, Any]]): A dict of static metadata for this asset.
            For example, users can provide information about the database table this
            asset corresponds to.
        skippable (bool): Whether this asset can be omitted during materialization, causing downstream
            dependencies to skip.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If
            not provided, the name "default" is used.
        code_version (Optional[str]): The version of the code for this specific asset,
            overriding the code version of the materialization function
        freshness_policy (Optional[FreshnessPolicy]): (Deprecated) A policy which indicates how up
            to date this asset is intended to be.
        auto_materialize_policy (Optional[AutoMaterializePolicy]): AutoMaterializePolicy to apply to
            the specified asset.
        backfill_policy (Optional[BackfillPolicy]): BackfillPolicy to apply to the specified asset.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the asset. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
        partitions_def (Optional[PartitionsDefinition]): Defines the set of partition keys that
            compose the asset.
    """

    def __new__(
        cls,
        key: CoercibleToAssetKey,
        *,
        deps: Optional[Iterable["CoercibleToAssetDep"]] = None,
        description: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        skippable: bool = False,
        group_name: Optional[str] = None,
        code_version: Optional[str] = None,
        freshness_policy: Optional[FreshnessPolicy] = None,
        automation_condition: Optional[AutomationCondition] = None,
        owners: Optional[Sequence[str]] = None,
        tags: Optional[Mapping[str, str]] = None,
        # TODO: FOU-243
        auto_materialize_policy: Optional[AutoMaterializePolicy] = None,
        partitions_def: Optional[PartitionsDefinition] = None,
    ):
        from dagster._core.definitions.asset_dep import coerce_to_deps_and_check_duplicates

        key = AssetKey.from_coercible(key)
        asset_deps = coerce_to_deps_and_check_duplicates(deps, key)

        validate_group_name(group_name)

        owners = check.opt_sequence_param(owners, "owners", of_type=str)
        for owner in owners:
            validate_asset_owner(owner, key)

        kind_tags = {tag_key for tag_key in (tags or {}).keys() if tag_key.startswith(KIND_PREFIX)}
        if kind_tags is not None and len(kind_tags) > 2:
            raise DagsterInvalidDefinitionError("Assets can have at most two kinds currently.")

        return super().__new__(
            cls,
            key=key,
            deps=asset_deps,
            description=check.opt_str_param(description, "description"),
            metadata=check.opt_mapping_param(metadata, "metadata", key_type=str),
            skippable=check.bool_param(skippable, "skippable"),
            group_name=check.opt_str_param(group_name, "group_name"),
            code_version=check.opt_str_param(code_version, "code_version"),
            freshness_policy=check.opt_inst_param(
                freshness_policy,
                "freshness_policy",
                FreshnessPolicy,
            ),
            automation_condition=check.opt_inst_param(
                resolve_automation_condition(automation_condition, auto_materialize_policy),
                "automation_condition",
                AutomationCondition,
            ),
            owners=owners,
            tags=validate_tags_strict(tags) or {},
            partitions_def=check.opt_inst_param(
                partitions_def, "partitions_def", PartitionsDefinition
            ),
        )

    @staticmethod
    def dagster_internal_init(
        *,
        key: CoercibleToAssetKey,
        deps: Optional[Iterable["CoercibleToAssetDep"]],
        description: Optional[str],
        metadata: Optional[Mapping[str, Any]],
        skippable: bool,
        group_name: Optional[str],
        code_version: Optional[str],
        freshness_policy: Optional[FreshnessPolicy],
        automation_condition: Optional[AutomationCondition],
        owners: Optional[Sequence[str]],
        tags: Optional[Mapping[str, str]],
        auto_materialize_policy: Optional[AutoMaterializePolicy],
        partitions_def: Optional[PartitionsDefinition],
    ) -> "AssetSpec":
        check.invariant(auto_materialize_policy is None)
        return AssetSpec(
            key=key,
            deps=deps,
            description=description,
            metadata=metadata,
            skippable=skippable,
            group_name=group_name,
            code_version=code_version,
            freshness_policy=freshness_policy,
            automation_condition=automation_condition,
            owners=owners,
            tags=tags,
            partitions_def=partitions_def,
        )

    @cached_property
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]:
        return {
            dep.asset_key: dep.partition_mapping
            for dep in self.deps
            if dep.partition_mapping is not None
        }

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        # TODO: FOU-243
        return (
            self.automation_condition.as_auto_materialize_policy()
            if self.automation_condition
            else None
        )

    @cached_property
    def kinds(self) -> Set[str]:
        return {tag[len(KIND_PREFIX) :] for tag in self.tags if tag.startswith(KIND_PREFIX)}
